/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.hotmail.frojasg1.applications.common.components.resizecomp;

import com.hotmail.frojasg1.applications.common.components.internationalization.InternException;
import com.frojasg1.general.number.IntegerFunctions;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.Rectangle;
import javax.swing.JLayeredPane;
import javax.swing.JPanel;
import javax.swing.JRootPane;
import javax.swing.JSplitPane;

/**
 *
 * @author Usuario
 */
public class ResizeRelocateItem
{
	// Flags:
	public static final int RESIZE_TO_RIGHT = 1;
	public static final int MOVE_TO_RIGHT = 2;
	public static final int RESIZE_TO_BOTTOM = 4;
	public static final int MOVE_TO_BOTTOM = 8;
	public static final int MOVE_LEFT_SIDE_PROPORTIONAL = 16;
	public static final int MOVE_RIGHT_SIDE_PROPORTIONAL = 32;
	public static final int MOVE_RIGHT_SIDE_TO_RIGHT = 64;
	public static final int LEAVE_WIDTH_UNALTERED = 128;
	public static final int MOVE_TOP_SIDE_PROPORTIONAL = 256;
	public static final int MOVE_BOTTOM_SIDE_PROPORTIONAL = 512;
	public static final int MOVE_BOTTOM_SIDE_TO_BOTTOM = 1024;
	public static final int LEAVE_HEIGHT_UNALTERED = 2048;

	public static final int MASK = 2 * LEAVE_HEIGHT_UNALTERED - 1;

	// For the calculations of the bounds of JSplitPane in calculateBounds_JSplitPane. options for parameter: int position
	public static final int RIGHT = 1;
	public static final int LEFT = 2;
	public static final int TOP = 3;
	public static final int BOTTOM = 4;
	
	protected ResizeRelocateItem_parent _parent = null;
	protected Component _component;
	protected int		_flags;
	
	protected int _originalWidthOfParent = -1;
	protected int _originalHeightOfParent = -1;

	protected int		_pixelsLeftFromTheComponentRightBoundaryToTheRightOfTheParent = -1;		// it is needed if RESIZE_TO_RIGHT flag is active
	protected int		_pixelsFromTheComponentLeftBoundaryToTheRightOfTheParent = -1;			// it is needed if MOVE_TO_RIGHT flag is active
	protected int		_pixelsLeftFromTheComponentBottomBoundaryToTheBottomOfTheParent = -1;	// it is needed if RESIZE_TO_BOTTOM flag is active
	protected int		_pixelsFromTheComponentUpperBoundaryToTheBottomOfTheParent = -1;			// it is needed if MOVE_TO_BOTTOM flag is active

	protected int _originalWidth = -1;
	protected int _originalHeight = -1;

	public ResizeRelocateItem( Component comp, int flags, ResizeRelocateItem_parent parent ) throws InternException
	{
		_parent = parent;
		setFlags( flags );
		initializeWithComponent( comp );
	}

	public void setFlags( int flags ) throws InternException
	{
		if( ( flags - ( flags & MASK ) ) != 0 )
			throw( new InternException( "Flags out of bounds" ) );
		
		if( isFlagActive( RESIZE_TO_RIGHT ) && isFlagActive( MOVE_TO_RIGHT ) )
			throw( new InternException( "Invalid flags: RESIZE_TO_RIGHT and MOVE_TO_RIGHT flags are both active" ) );

		if( isFlagActive( RESIZE_TO_BOTTOM ) && isFlagActive( MOVE_TO_BOTTOM ) )
			throw( new InternException( "Invalid flags: RESIZE_TO_BOTTOM and MOVE_TO_BOTTOM flags are both active" ) );

		if( ( isFlagActive( RESIZE_TO_RIGHT ) || isFlagActive( MOVE_TO_RIGHT ) ) &&
			isFlagActive( MOVE_LEFT_SIDE_PROPORTIONAL ) )
			throw( new InternException( "Invalid flags: MOVE_LEFT_SIDE_PROPORTIONAL is incompatible with " +
										"RESIZE_TO_RIGHT and MOVE_TO_RIGHT flags" ) );

		if( ( isFlagActive( RESIZE_TO_BOTTOM ) || isFlagActive( MOVE_TO_BOTTOM ) ) &&
			isFlagActive( MOVE_TOP_SIDE_PROPORTIONAL ) )
			throw( new InternException( "Invalid flags: MOVE_TOP_SIDE_PROPORTIONAL is incompatible with " +
										"RESIZE_TO_BOTTOM and MOVE_TO_BOTTOM flags" ) );

		if( isFlagActive( MOVE_LEFT_SIDE_PROPORTIONAL ) &&
			!( isFlagActive( MOVE_RIGHT_SIDE_PROPORTIONAL ) ||
				isFlagActive( MOVE_RIGHT_SIDE_TO_RIGHT ) ||
				isFlagActive( LEAVE_WIDTH_UNALTERED ) ) )
			throw( new InternException( "Invalid flags: MOVE_LEFT_SIDE_PROPORTIONAL must be combined with " +
										"MOVE_RIGHT_SIDE_PROPORTIONAL or MOVE_RIGHT_SIDE_TO_RIGHT or LEAVE_WIDTH_UNALTERED flags" ) );

		if( isFlagActive( MOVE_TOP_SIDE_PROPORTIONAL ) &&
			!( isFlagActive( MOVE_BOTTOM_SIDE_PROPORTIONAL ) ||
				isFlagActive( MOVE_BOTTOM_SIDE_TO_BOTTOM ) ||
				isFlagActive( LEAVE_HEIGHT_UNALTERED ) ) )
			throw( new InternException( "Invalid flags: MOVE_TOP_SIDE_PROPORTIONAL must be combined with " +
										"MOVE_BOTTOM_SIDE_PROPORTIONAL or MOVE_BOTTOM_SIDE_TO_BOTTOM or LEAVE_HEIGHT_UNALTERED flags" ) );

/*	This two posibilities are now possible.
		if( !isFlagActive( MOVE_LEFT_SIDE_PROPORTIONAL ) &&
			( isFlagActive( MOVE_RIGHT_SIDE_PROPORTIONAL ) ||
				isFlagActive( MOVE_RIGHT_SIDE_TO_RIGHT ) ||
				isFlagActive( LEAVE_WIDTH_UNALTERED ) ) )
			throw( new InternException( "Invalid flags: MOVE_RIGHT_SIDE_PROPORTIONAL, MOVE_RIGHT_SIDE_TO_RIGHT and " +
										"LEAVE_WIDTH_UNALTERED flags can only be active in combination with" +
										"MOVE_LEFT_SIDE_PROPORTIONAL flag." ) );

		if( !isFlagActive( MOVE_TOP_SIDE_PROPORTIONAL ) &&
			( isFlagActive( MOVE_BOTTOM_SIDE_PROPORTIONAL ) ||
				isFlagActive( MOVE_BOTTOM_SIDE_TO_BOTTOM ) ||
				isFlagActive( LEAVE_HEIGHT_UNALTERED ) ) )
			throw( new InternException( "Invalid flags: MOVE_BOTTOM_SIDE_PROPORTIONAL, MOVE_BOTTOM_SIDE_TO_BOTTOM and " +
										"LEAVE_HEIGHT_UNALTERED flags can only be active in combination with" +
										"MOVE_TOP_SIDE_PROPORTIONAL flag." ) );
*/
		if( isFlagActive( MOVE_RIGHT_SIDE_PROPORTIONAL ) &&	isFlagActive( MOVE_RIGHT_SIDE_TO_RIGHT ) ||
			isFlagActive( MOVE_RIGHT_SIDE_PROPORTIONAL ) &&	isFlagActive( LEAVE_WIDTH_UNALTERED )  ||
			isFlagActive( MOVE_RIGHT_SIDE_TO_RIGHT ) &&	isFlagActive( LEAVE_WIDTH_UNALTERED ) )
			throw( new InternException( "Invalid flags: only one of these flags can be active at the same time: " +
										"MOVE_RIGHT_SIDE_PROPORTIONAL, MOVE_RIGHT_SIDE_TO_RIGHT, LEAVE_WIDTH_UNALTERED" ) );

		if( isFlagActive( MOVE_BOTTOM_SIDE_PROPORTIONAL ) &&	isFlagActive( MOVE_BOTTOM_SIDE_TO_BOTTOM ) ||
			isFlagActive( MOVE_BOTTOM_SIDE_PROPORTIONAL ) &&	isFlagActive( LEAVE_HEIGHT_UNALTERED )  ||
			isFlagActive( MOVE_BOTTOM_SIDE_TO_BOTTOM ) &&	isFlagActive( LEAVE_HEIGHT_UNALTERED ) )
			throw( new InternException( "Invalid flags: only one of these flags can be active at the same time: " +
										"MOVE_BOTTOM_SIDE_PROPORTIONAL, MOVE_BOTTOM_SIDE_TO_BOTTOM, LEAVE_HEIGHT_UNALTERED" ) );

		_flags = flags;
	}
	
	public void setParent( ResizeRelocateItem_parent parent )
	{
		_parent = parent;
	}

	public void initializeWithComponent( Component comp ) throws InternException
	{
		_component = comp;

		if( _component == null )
			throw( new InternException( "Component null" ) );

		Component parent = _component.getParent();
		Rectangle componentBounds = getComponentBounds( comp );
		Rectangle parentBounds = getParentBounds( comp );

		if( parent == null )
			throw( new InternException( "Parent component null" ) );

		_originalWidth = _component.getWidth();
		_originalHeight = _component.getHeight();

		_originalWidthOfParent = parent.getWidth();
		_originalHeightOfParent = parent.getHeight();


//		if( isFlagActive( RESIZE_TO_RIGHT ) )
		{
			_pixelsLeftFromTheComponentRightBoundaryToTheRightOfTheParent = (int) ( parentBounds.getWidth() -
																					( componentBounds.getX() - parentBounds.getX() ) -
																					componentBounds.getWidth()
/*																					( componentBounds.getWidth() < parentBounds.getWidth() ?
																						componentBounds.getWidth() :
																						parentBounds.getWidth()
																						) */
																					);
		}
//		else if( isFlagActive( MOVE_TO_RIGHT ) )
		{
			_pixelsFromTheComponentLeftBoundaryToTheRightOfTheParent =	(int) ( parentBounds.getWidth() - parentBounds.getX() -
																				componentBounds.getX() );
		}
		
//		if( isFlagActive( RESIZE_TO_BOTTOM ) )
		{
			_pixelsLeftFromTheComponentBottomBoundaryToTheBottomOfTheParent =	(int) ( parentBounds.getHeight() -
																						(componentBounds.getY() - parentBounds.getY()) -
																						componentBounds.getHeight()
/*																						( componentBounds.getHeight() < parentBounds.getHeight() ?
																							componentBounds.getHeight() :
																							parentBounds.getHeight()
																							)*/
																						);
		}
//		else if( isFlagActive( MOVE_TO_BOTTOM ) )
		{
			_pixelsFromTheComponentUpperBoundaryToTheBottomOfTheParent =	(int) ( parentBounds.getHeight() -
																					(componentBounds.getY() - parentBounds.getY())
																					);
		}
	}
	
	public boolean isFlagActive( int flag )
	{
		return( (_flags & flag) != 0 );
	}

	public void execute()
	{
		Rectangle parentBounds = getParentBounds( _component );
		Rectangle newBounds = calculateNewBounds( parentBounds );
		_component.setBounds(newBounds);
		
		if( _component instanceof ResizableComponentInterface )
		{
			ResizableComponentInterface rci = (ResizableComponentInterface) _component;
			rci.doTasksAfterResizingComponent();
		}
	}

	public Rectangle calculateNewBounds( Rectangle parentBounds )
	{
		Rectangle newBounds = getComponentBounds( _component );
//		Rectangle newBounds = _component.getBounds();

		Component parent = _component.getParent();

		if( parent != null )
		{
			if( parent instanceof JSplitPane )
			{
				_component.setLocation( parentBounds.x, parentBounds.y );
			}
/*
			int newX = (int) _component.getX();
			int newY = (int) _component.getY();
			int width = (int) _component.getWidth();
			int height = (int) _component.getHeight();
*/
			int newX = (int) newBounds.getX();
			int newY = (int) newBounds.getY();
			int width = (int) newBounds.getWidth();
			int height = (int) newBounds.getHeight();

			if( isFlagActive( RESIZE_TO_RIGHT ) )
			{
				int delta = (int)(	parentBounds.getWidth() -
									( newBounds.getX() - parentBounds.getX() ) -
									newBounds.getWidth() ) -
									_pixelsLeftFromTheComponentRightBoundaryToTheRightOfTheParent;
				if( delta != 0 )
				{
					width = (int) Math.max( 
										Math.min( newBounds.getWidth() + delta, _component.getMaximumSize().getWidth() ),
										_component.getMinimumSize().getWidth() );
//					_component.setSize( width, (int) _component.getHeight() );
				}
			}
			else if( isFlagActive( MOVE_TO_RIGHT ) )
			{
				int delta = (int) ( parentBounds.getWidth() -
									( newBounds.getX() - parentBounds.getX() ) ) -
									_pixelsFromTheComponentLeftBoundaryToTheRightOfTheParent;
				if( delta != 0 )
				{
					newX = (int) newBounds.getX() + delta;
//					_component.setLocation( newX, (int) _component.getY() );
				}
			}
			else
			{
				if( _originalWidthOfParent != 0 )
				{
					double factor = (double) parentBounds.getWidth() / _originalWidthOfParent;
					
					if( isFlagActive( MOVE_LEFT_SIDE_PROPORTIONAL ) )
						newX = (int)( parentBounds.getWidth() - _pixelsFromTheComponentLeftBoundaryToTheRightOfTheParent * factor );

					int rightBound = newX + _component.getWidth();

					if( isFlagActive( MOVE_RIGHT_SIDE_PROPORTIONAL ) )
					{
						rightBound = (int)( parentBounds.getWidth() - _pixelsLeftFromTheComponentRightBoundaryToTheRightOfTheParent * factor );
					}
					else if( isFlagActive( MOVE_RIGHT_SIDE_TO_RIGHT ) )
					{
						rightBound = (int)( parentBounds.getWidth() - _pixelsLeftFromTheComponentRightBoundaryToTheRightOfTheParent );
					}
					else if( isFlagActive( LEAVE_WIDTH_UNALTERED ) )
					{
						rightBound = newX + _originalWidth;
					}

					width = rightBound - newX;
				}
			}

			if( isFlagActive( RESIZE_TO_BOTTOM ) )
			{
				int delta = (int) ( parentBounds.getHeight() -
									( newBounds.getY()  - parentBounds.getY()) -
									newBounds.getHeight() ) -
									_pixelsLeftFromTheComponentBottomBoundaryToTheBottomOfTheParent;
				if( delta != 0 )
				{
					height = (int) Math.max(
									Math.min( newBounds.getHeight() + delta, _component.getMaximumSize().getHeight() ),
									_component.getMinimumSize().getHeight() );
//					_component.setSize( new Double( _component.getWidth() ).intValue(), new Double( _component.getHeight() ).intValue() + delta );
				}
			}
			else if( isFlagActive( MOVE_TO_BOTTOM ) )
			{
				int delta = (int) ( parentBounds.getHeight() -
									( newBounds.getY()  - parentBounds.getY() ) ) -
									_pixelsFromTheComponentUpperBoundaryToTheBottomOfTheParent;
				if( delta != 0 )
				{
					newY = (int)newBounds.getY() + delta;
//					_component.setLocation( (int) _component.getX(), newY );
				}
			}
/*			else
			{
				if( _originalHeightOfParent != 0 )
				{
					double factor = (double) parentBounds.getWidth() / _originalHeightOfParent;
					
					if( isFlagActive( MOVE_TOP_SIDE_PROPORTIONAL ) )
						newY = (int)( parentBounds.getHeight() - _pixelsFromTheComponentLeftBoundaryToTheRightOfTheParent * factor );

					int bottomBound = newY + _component.getHeight();

					if( isFlagActive( MOVE_RIGHT_SIDE_PROPORTIONAL ) )
					{
						bottomBound = (int)( parentBounds.getHeight() - _pixelsLeftFromTheComponentBottomBoundaryToTheBottomOfTheParent * factor );
					}
					else if( isFlagActive( MOVE_RIGHT_SIDE_TO_RIGHT ) )
					{
						bottomBound = (int)( parentBounds.getHeight() - _pixelsLeftFromTheComponentBottomBoundaryToTheBottomOfTheParent );
					}
					else if( isFlagActive( LEAVE_WIDTH_UNALTERED ) )
					{
						bottomBound = newX + _originalHeight;
					}

					height = bottomBound - newY;
				}
			}
*/
			Dimension minSize = _component.getMinimumSize();
			Dimension maxSize = _component.getMaximumSize();
			
			width = (int) Math.min( maxSize.getWidth(), Math.max( minSize.getWidth(), width ) );
			height = (int) Math.min( maxSize.getHeight(), Math.max( minSize.getHeight(), height ) );

			newBounds = new Rectangle( newX, newY, width, height );
		}
		return( newBounds );
	}

	protected Rectangle getParentBounds( Component comp ) throws IllegalArgumentException
	{
		Component parent = comp.getParent();
/*
		if( ( parent instanceof JPanel ) &&
			( parent.getName() != null ) &&
			( parent.getName().length() >= 5 ) &&
			parent.getName().substring( 0, 5 ).equals( "null." ) )
		{
			JPanel panel = (JPanel) parent;
			panel.revalidate();
			panel.repaint();
		}
*/

		// if the parent is a kind of root pane, we skip all of them until we reach the root component.
		while( ( parent != null ) && (parent.getParent() != null ) &&
				( ( parent instanceof JLayeredPane ) || ( parent instanceof JRootPane ) ||
					( parent instanceof JPanel ) && ( parent.getName() != null ) &&
					( parent.getName().length() >= 5 ) &&
					parent.getName().substring( 0, 5 ).equals( "null." )
				)
			)
		{
			parent = parent.getParent();
		}

		Rectangle result = null;
		if( parent != null )
		{
			int left = -1;
			int top = -1;
			int width = -1;
			int height = -1;

			if( parent instanceof JSplitPane )
			{
				JSplitPane jsp = (JSplitPane) parent;

				int dividerLocation = jsp.getDividerLocation();
				Rectangle bounds = jsp.getBounds();

				if( jsp.getOrientation() == JSplitPane.HORIZONTAL_SPLIT )
				{
					if( jsp.getLeftComponent() == comp )		result = calculateBounds_JSplitPane( jsp, bounds, dividerLocation, LEFT );
					else if( jsp.getRightComponent() == comp )	result = calculateBounds_JSplitPane( jsp, bounds, dividerLocation, RIGHT );
					else
					{
						throw( new IllegalArgumentException( "The split pane parent was HorizontalSplit, but the child was not found" ) );
					}
				}
				else if( jsp.getOrientation() == JSplitPane.VERTICAL_SPLIT )
				{
					if( jsp.getTopComponent() == comp )		result = calculateBounds_JSplitPane( jsp, bounds, dividerLocation, TOP );
					else if( jsp.getBottomComponent() == comp )	result = calculateBounds_JSplitPane( jsp, bounds, dividerLocation, BOTTOM );
					else
					{
						throw( new IllegalArgumentException( "The split pane parent was VerticalSplit, but the child was not found" ) );
					}
				}
				else
				{
					throw( new IllegalArgumentException( "The split pane parent was neither HorizontalSplit neither VerticalSplit" ) );
				}
			}
			else
			{
//				Dimension dim = parent.getSize();
				Dimension dim = getComponentSize( parent );
				left = 0;
				top = 0;
				width = dim.width;
				height = dim.height;

				result = new Rectangle( left, top, width, height );
			}
		}
		return( result );
	}

	protected Dimension getComponentSize( Component comp )
	{
		Dimension dim = comp.getSize();

		int width = dim.width;
		int height = dim.height;

		if( width == 0 )	width = (int) comp.getPreferredSize().getWidth();
		if( height == 0 )	height = (int) comp.getPreferredSize().getHeight();

		Dimension result = new Dimension( width, height );
		
		return( result );
	}

	protected Rectangle getComponentBounds( Component comp )
	{
		Rectangle result = _component.getBounds();
		
		Dimension size = getComponentSize( comp );
		
		result.width = size.width;
		result.height = size.height;

		return( result );
	}

	protected Rectangle calculateBounds_JSplitPane( JSplitPane jsp, Rectangle newBounds, int dividerLocation, int position )
	{
		Insets insets = jsp.getInsets();

		int left = -1;
		int top = -1;
		int width = -1;
		int height = -1;

		if( position == LEFT )
		{
			top = insets.top;
			height = (int) newBounds.getHeight() - top - insets.bottom;
			left = insets.left;
			width = dividerLocation - left;
		}
		else if( position == RIGHT )
		{
			top = insets.top;
			height = (int) newBounds.getHeight() - top - insets.bottom;
			left = dividerLocation + jsp.getDividerSize();
			width = (int) newBounds.getWidth() - left - insets.right;
		}
		else if( position == TOP )
		{
			left = insets.left;
			width = (int)newBounds.getWidth() - left - insets.right;
			top = insets.top;
			height = dividerLocation - top;
		}
		else if( position == BOTTOM )
		{
			left = insets.left;
			width = (int)newBounds.getWidth() - left - insets.right;
			top = dividerLocation + jsp.getDividerSize();
			height = (int) newBounds.getHeight() - top - insets.bottom;
		}

		Rectangle result = new Rectangle( left, top, width, height );
		return( result );
	}
	
	public static ResizeRelocateItem buildResizeRelocateItem( Component comp, int flags, ResizeRelocateItem_parent parent ) throws InternException
	{
		ResizeRelocateItem result = null;
		
		if( comp instanceof JSplitPane )
		{
			result = new ResizeRelocateItem_JSplitPane( (JSplitPane) comp, flags, parent );
		}
		else
		{
			result = new ResizeRelocateItem( comp, flags, parent );
		}
		
		return( result );
	}

	
}
